-- Paolo Torrini, Verimag, 2007

module TopoTree where

import Prelude
import IO
import Debug.Trace
import List
import VVlib

--------- AND-OR trees

data Gate = GAnd | GOr 
  deriving (Ord, Eq, Show) 

data Nd a = Nd { gate :: Gate,
                 ants :: [a]
               } 
  deriving (Ord, Eq, Show)

data Tr a = Tr { root :: a,
                 brans :: Nd (Tr a)
               }
  deriving (Ord, Eq, Show)

---------- Logic

data Lg a = AND [Lg a] | OR [Lg a] | IMP (Lg a) (Lg a) | NOT (Lg a) 
            | FF | TT | CTR a | OBS a | AOBS a
 deriving (Ord, Eq, Show)

---------- Fault representation

data Info = Y | N | I
  deriving (Ord, Eq, Show)

data Fact = Place { info :: Info,
                    name :: String 
                  }
          | Trans { info :: Info,
                    name :: String 
                  } 
  deriving (Ord, Eq, Show)

data Event = Event { place :: Fact,
                     trans :: Fact
                   }
  deriving (Ord, Eq, Show)

type FTree = Tr Event

class Ev a where
  isWkContr :: a -> Bool
  isStContr :: a -> Bool
  isObs     :: a -> Bool

instance Ev Event where
 isWkContr (Event _ t) = case t of 
   Trans Y _ -> True
   Trans I _ -> True
   _ -> False
 isStContr (Event _ t) = case t of 
   Trans Y _ -> True
   _         -> False
 isObs (Event p _) = case p of
   Place N _ -> False
   _         -> True

--------- Event functions

mkPl :: Info -> String -> Event
mkPl i s = Event (Place i s) (Trans N $ s ++ ":d")

mkTr :: Info -> String -> Event
mkTr i s = Event (Place N $ s ++ ":d") (Trans i s)

mkEv :: Info -> String -> Info -> String -> Event
mkEv i r j s = Event (Place i r) (Trans j s)

mkAnd :: Event -> [FTree] -> FTree
mkAnd x ls = Tr x (Nd GAnd ls)

mkOr :: Event -> [FTree] -> FTree 
mkOr x ls = Tr x (Nd GOr ls)

--------- auxiliary

getLeaves :: Tr a -> [a]
getLeaves (Tr x (Nd w ls)) = case ls of
   [] -> [x]      
   _ -> concat $ map getLeaves ls

getContrLeaves :: Ev a => Tr a -> [a]
getContrLeaves ls = filter isStContr $ getLeaves ls 

convTr :: (a -> b) -> Tr a -> Tr b
convTr f (Tr x (Nd w ls)) = 
   Tr (f x) $ Nd w $ map (convTr f) ls
 
--------- Properties of Boolean functions over trees

trueRoot :: (a -> Bool) -> Tr a -> Bool
trueRoot f (Tr x _) = f x

nextOrRoot :: (Tr a -> Bool) -> Tr a -> Bool
nextOrRoot f (Tr _ (Nd w ls)) = genOr f ls 

nextAndRoot :: (Tr a -> Bool) -> Tr a -> Bool
nextAndRoot f (Tr _ (Nd w ls)) = genAnd f ls 

diamonRoot :: (Tr a -> Bool) -> Tr a -> Bool
diamonRoot f t = case f t of
  True -> True
  _ -> nextOrRoot (diamonRoot f) t 

boxRoot :: (Tr a -> Bool) -> Tr a -> Bool
boxRoot f t = case f t of
  False -> False
  _ -> nextAndRoot (boxRoot f) t 

safeRoot :: Show a => (Tr a -> Bool) -> Tr a -> Bool
safeRoot f t = case f t of
  True -> False
  _    -> nextAndRoot (safeRoot f) t 
-- safeRoot f = boxRoot (negP f) = not diamonRoot f

-- not needed
evntallyRoot :: (Tr a -> Bool) -> Tr a -> Bool
evntallyRoot f t = case f t of
    True -> nextAndRoot (evntallyRoot f) t
    _ -> case (ants $ brans t) of
       [] -> False
       _ -> nextAndRoot (evntallyRoot f) t

-- not needed
reachTr :: (Tr a -> Bool) -> Tr a -> Bool
reachTr f t = case f t of
    True -> True
    _ -> case (ants $ brans t) of
        [] -> False
        _ -> nextAndRoot (evntallyRoot f) t
-- evntallyRoot f = reachTr (boxRoot f) 

-- not needed
liveTr :: (Tr a -> Bool) -> Tr a -> Bool
liveTr f t = case reachTr f t of
    True -> nextAndRoot (liveTr f) t
    _ -> False  
-- liveTr f = boxRoot (reachTr f)

nextFilter :: Eq a => (a -> Bool) -> Tr a -> Tr a
nextFilter f (Tr x (Nd w ls)) = case w of 
   GAnd -> Tr x (Nd w (filter (f . root) (map (nextFilter f) ls)))
   GOr  -> if (filter (f . root) ls == ls) then 
                Tr x (Nd w (map (nextFilter f) ls))
           else Tr x (Nd GAnd []) 
 
---------------- Highlight tree regions

setA :: (Tr a -> Bool) -> (Bool -> a -> b) -> Tr a -> Tr b
setA f g t = case t of
   Tr x (Nd w ls) -> Tr (g (f t) x) (Nd w $ map (setA f g) ls)

closA :: (Show a, Show b) => 
           (Tr a -> Bool) -> (Bool -> a -> b) -> Tr a -> Tr b
closA f g t = case t of
   Tr x (Nd w ls) -> if (safeRoot f t) then convTr (g False) t
       else Tr (g True x) $ Nd w $ map (closA f g) ls
-- less efficient: closA f g t = case t of
--  Tr x (Nd w ls) -> Tr (g $ diamonRoot f t) (Nd w $ map (closA f g) ls)

neighA :: (Tr a -> Bool) -> (Bool -> a -> b) -> Tr a -> Tr b
neighA f g t = case t of 
    Tr x (Nd w ls) -> if (f t) then convTr (g True) t
         else (Tr (g False x) $ Nd w $ map (neighA f g) ls)
-- less efficient: neighA f g t = case t of 
--  Tr x (Nd w ls) -> Tr (g (f t) x) $ map (neighA f g) ls

notA :: (Show a, Show b) => 
          (Tr a -> Bool) -> (Bool -> a -> b) -> Tr a -> Tr b
notA f g t = neighA (safeRoot f) g t
-- notA f g t = case t of 
--    Tr x (Nd w ls) -> if (safeRoot f t) then convTr (g True) t
--        else Tr (g False x) $ Nd w $ map (notA f g) ls
-- less efficient: notA f g t = case t of
--  Tr x (Nd w ls) -> Tr (g (f t) x) (Nd w $ map (notA f g) ls)

evntA :: (Tr a -> Bool) -> (Bool -> a -> b) -> Tr a -> Tr b
evntA f g t = neighA (evntallyRoot f) g t
-- evntA f g t = case t of 
--    Tr x (Nd w ls) -> if (evntallyRoot f t) then convTr (g True) t
--        else Tr (g False x) $ Nd w $ map (evntA f g) ls
-- less efficient: evntA f g t = case t of
--  Tr x (Nd w ls) -> Tr (g $ evntallyRoot f t) (Nd w $ map (evntA f g) ls)
-- also: notA (\x -> not (f x)) g t

intrA :: (Tr a -> Bool) -> (Bool -> a -> b) -> Tr a -> Tr b
intrA f g t = neighA (boxRoot f) g t
-- intrA f  g t = case t of 
--    Tr x (Nd w ls) -> if (boxRoot f t) then convTr (g True) t
--        else Tr (g False x) $ Nd w $ map (intrA f g) ls
-- less efficient: intrA f g t = case t of
--  Tr x (Nd w ls) -> Tr (g $ boxRoot f t) (Nd w $ map (intrA f g) ls)
         
conA :: [Tr a -> Bool] -> (Bool -> a -> b) -> Tr a -> Tr b
conA fs g t = neighA (funAnd fs) g t

orA :: [Tr a -> Bool] -> (Bool -> a -> b) -> Tr a -> Tr b
orA fs g t = neighA (funOr fs) g t


------------ Cut out regions

closBC :: Show a => (Tr a -> Bool) -> Tr a -> [Tr a]  -- max singl list
closBC f t = case t of
--   Tr x (Nd GXAnd ls) -> if (safeRoot f t) then [] 
--       else let xxs = [Tr y (Nd z $ concat $ map (closBC f) ms) |
--                           Tr y (Nd z ms) <- ls]
--            in [Tr x $ Nd GXAnd xxs]
   Tr x (Nd w ls) -> if (safeRoot f t) then [] 
       else [Tr x $ Nd w $ concat $ map (closBC f) ls]

pruneBC :: Show a => (Tr a -> Bool) -> Tr a -> [Tr a]  -- max singl list
pruneBC f t = let xs = closBC f t 
  in map (cutB f) xs

cutB :: (Tr a -> Bool) -> Tr a -> Tr a
cutB f t = case t of 
     Tr x (Nd w ls) -> if f t then Tr x (Nd w [])
          else Tr x (Nd w $ map (cutB f) ls) 

closB :: Show a => (Tr a -> Bool) -> Tr a -> [Tr a]  -- max singl list
closB f t = case t of
   Tr x (Nd w ls) -> if (safeRoot f t) then [] 
       else [Tr x $ Nd w $ concat $ map (closB f) ls]

pruneB :: Show a => (Tr a -> Bool) -> Tr a -> [Tr a]  -- max singl list
pruneB f t = let xs = closB f t 
  in map (cutB f) xs

neighB :: (Tr a -> Bool) -> Tr a -> [Tr a]
neighB f t = case t of 
    Tr x (Nd w ls) -> if (f t) then [t]
         else concat $ map (neighB f) ls

notB :: Show a => (Tr a -> Bool) -> Tr a -> [Tr a]
notB f t = neighB (safeRoot f) t

evntB :: (Tr a -> Bool) -> Tr a -> [Tr a]
evntB f t = neighB (evntallyRoot f) t

intrB :: (Tr a -> Bool) -> Tr a -> [Tr a]
intrB f t = neighB (boxRoot f) t

conB :: [Tr a -> Bool] -> Tr a -> [Tr a]
conB fs t = neighB (funAnd fs) t

orB :: [Tr a -> Bool] -> Tr a -> [Tr a]
orB fs t = neighB (funOr fs) t

-------- Translation functions

treeNames :: FTree -> [String]
treeNames t = let 
     Event x y = root t 
     Nd _ ls   = brans t
     w         = name x
     z         = name y
  in nub $ concat [[w,z],(concat $ map treeNames ls)]  

---------- Fault tree functions

ft2lg :: FTree -> [Lg String]
ft2lg t = let 
    ws = pruneTree t
    zs = states [[root ws]] [ws]
    cl = getContrLeaves ws
    xs = selectStates zs cl 
  in trace ("\n TREE (FTree style): \n" ++ show t) $ 
     trace ("\n PRUNED TREE: \n" ++ show ws) $ 
     trace ("\n LEAF NODES: \n" ++ show (getLeaves t)) $ 
     trace ("\n CONTROLLABLE LEAF NODES: \n" ++ show cl) $ 
     trace ("\n STATES: \n" ++ show zs) $ nub xs  

pruneTreeX :: FTree -> [FTree]
pruneTreeX t = map (nextFilter isObs) $ pruneBC (trueRoot isStContr) t 

pruneTree :: FTree -> FTree
pruneTree t = case (pruneBC (trueRoot isWkContr) (preProcFT t)) of
  [x] -> nextFilter isObs x
  _   -> error "TopoTree: pruneTree - not a tree"

-- arg 1 is the input state set, arg 2 are the trees and give 
-- the transitions, output is a state set 
states :: (Show a, Eq a) => [[a]] -> [Tr a] -> [[a]]
states zss ts = let
     zss1 = nub zss
     ts1  = nub ts
 in case ts1 of
       _  -> nub $ concat [zss1,
          concat [transAppl zss1 (delete t ts1) t | t <- ts1]]

transAppl :: (Show a, Eq a) => [[a]] -> [Tr a] -> Tr a -> [[a]]
transAppl zss rs t@(Tr x (Nd w ls)) = let 
      hs  = nub $ map root ls
      kss = nub $ case w of 
             GAnd -> [(nub $ andReplace ws x hs) | ws <- zss]
             GOr  -> nub $ concat [(nub $ orReplace ws x hs) | ws <- zss]
   in case ls of
      [] -> zss
      _  -> states kss (concat [ls, rs])

-- arg 1 are states, arg 2 are control nodes,
-- return lists of formulas for controllable states
selectStates :: [[Event]] -> [Event] -> [Lg String]
selectStates ls cl = 
   if cl == [] then trace ("\n NO SAFETY CONDITIONS \n") $ []
   else let 
        hs = filter (interNE cl) ls  
        ks = map (transState cl) hs 
     in trace ("\n CONTROLLABLE STATES: \n" ++ show hs) $ 
        trace ("\n TRANSLATED STATES: \n" 
                       ++ show (concat ks) ++ "\n") $ 
        trace ("\n SAFETY CONDITIONS: \n") $
        map (AND . sort) (filter contrFS $ concat ks)

andReplace :: Eq a => [a] -> a -> [a] -> [a]
andReplace ls x ws = let 
     hs = delete x ls
  in if length ls == length hs 
     then ls else concat [hs,ws] 

orReplace :: Eq a => [a] -> a -> [a] -> [[a]]
orReplace ls x ws = let 
     hs = delete x ls
  in if length ls == length hs 
     then [ls] else [p:hs | p <- ws] 

contrFS :: [Lg String] -> Bool
contrFS ls = case ls of
   []        -> False
   (CTR _):_ -> True
   _:hs      -> contrFS hs 

transState :: [Event] -> [Event] -> [[Lg String]]
transState cl ls = cProd $ map (formState cl) ls 

formState :: [Event] -> Event -> [Lg String]
formState cl x = case elem x cl of
    True  -> [evalEv x, evalCd x] 
    False -> [evalCd x] 

evalEv :: Event -> Lg String
evalEv w@(Event _ (Trans i x)) = case i of 
   N -> trace (show w) $ error "TopoTree: evalEv, uncontrollable event (N)"
   I -> trace (show w) $ error "TopoTree: evalEv, uncontrollable event (I)"
   Y -> CTR x

evalCd :: Event -> Lg String
evalCd w@(Event (Place i x) _) = case i of
   N -> trace (show w) $ error "TopoTree: evalCd, unobservable condition"
   Y -> OBS x
   I -> AOBS x

trEq :: Event -> Event -> Bool
trEq (Event _ t1) (Event _ t2) = 
  if t1 == t2 then True else False

-- preprocess trees to cope with I transitions
preProcFT :: FTree -> FTree
preProcFT t = case t of
  Tr (Event k (Trans I x)) 
              (Nd GAnd []) -> Tr (Event (preprC k) (Trans I x)) 
                                                (Nd GAnd [])
  Tr (Event k (Trans I x)) 
              (Nd GAnd ls) -> 
       Tr (Event k (Trans N x)) 
              (Nd GAnd 
                   [  Tr (Event (preprC p) (Trans I ww)) (Nd GAnd []) 
                      | Tr (Event p (Trans _ ww)) _ <- ls  ]
          )
  Tr x (Nd w ls)           -> 
       Tr x (Nd w $ map preProcFT ls)

preprC :: Fact -> Fact
preprC x = case x of
   Place Y n -> Place I n
   _        -> x

-----------------------------------------------------------------

