-- Paolo Torrini, Verimag, 2007

module UDType where

import qualified Data.Map as Map
import Prelude
import IO
import Debug.Trace
import Data.Typeable
import Language.Haskell.Syntax
import VVlib
import List
import qualified TopoTree as T

import Algo1

--------------------------------------------------------------
------ UDG term types ----------------------------------------

type Graphterm = [Elem]

data Elem = L (ElemId, ElemProps)
          | R ElemId 
  deriving (Ord, Eq, Show, Typeable)

type ElemId = String

data ElemProps = N (TypeId, 
                    [Attribute], 
                    [Elem]) 
               | E (TypeId, 
                    [Attribute], 
                    Elem)
  deriving (Ord, Eq, Show, Typeable)

type TypeId = String

data Attribute = A (String, String) | M String
  deriving (Ord, Eq, Show, Typeable)

------------------------------------------------------------------
------ Conversion from HsExp to UDG types ------------------------

exp2graph :: HsExp -> Maybe Graphterm
exp2graph e = case e of 
   HsList ws -> mapM exp2elem ws
   _ -> trace (show e) $ Nothing

exp2elem :: HsExp -> Maybe Elem
exp2elem e = case e of
  HsApp x y -> case (extCon x, y) of
     (Just "L", HsTuple [a,b]) -> do 
                y1 <- exp2id a
                y2 <- exp2props b
                return $ L (y1, y2)
     (Just "R", y) -> do 
                y1 <- exp2id y
                return $ R y1 
     _ ->  trace (show e) $ Nothing
  HsParen w -> exp2elem w
  _ ->  trace (show e) $ Nothing

extCon :: HsExp -> Maybe String
extCon e = case e of
  HsCon (UnQual (HsIdent x)) -> Just x
  HsParen w -> extCon w
  _ ->  trace (show e) $ Nothing

exp2id :: HsExp -> Maybe String
exp2id e = case e of 
  HsLit (HsString x) -> Just x
  HsParen w -> exp2id w
  _ ->  trace (show e) $ Nothing

exp2props :: HsExp -> Maybe ElemProps
exp2props e = case e of
  HsApp x y -> case (extCon x, y) of
    (Just "N", HsTuple [a, HsList bs, HsList cs]) -> do
         y1 <- exp2id a
         xs <- mapM exp2att bs
         zs <- mapM exp2elem cs
         return $ N (y1, xs, zs) 
    (Just "E", HsTuple [a, HsList bs, c]) -> do
         y1 <- exp2id a
         xs <- mapM exp2att bs
         z <- exp2elem c
         return $ E (y1, xs, z) 
    _ ->  trace (show e) $ Nothing
  HsParen w -> exp2props w
  _ ->  trace (show e) $ Nothing

exp2att :: HsExp -> Maybe Attribute
exp2att e = case e of 
  HsApp x y -> case (extCon x, y) of
    (Just "A", HsTuple [a,b]) -> do 
              y1 <- exp2id a  
              y2 <- exp2id b
              return $ A (y1, y2)
    (Just "M", _) -> return $ M (show y)  
    _ ->  trace (show e) $ Nothing
  HsParen w -> exp2att w
  _ ->  trace (show e) $ Nothing

--------------------------------------------------------------------
-------- 1st UDG transformation ------------------------------------

-- set the correct Elem type (first arg of N and E)
repaint :: Elem -> Elem
repaint e = case e of
  R _ -> e
  L (m, xs) -> L (m, case xs of
     N (n, ys, ws) -> N (maybe n id $ tyRep ys, ys, map repaint ws) 
     E (n, ys, w)  -> E (maybe n id $ tyRep ys, ys, repaint w)) 

tyRep :: [Attribute] -> Maybe TypeId
tyRep ls = do hs <- extAttr ls "OBJECT"
              ks <- return $ dropWhile (\x -> x == ' ') hs
              return $ takeWhile (\x -> x /= ' ' && x /= ':') ks

--------------------------------------------------------------------
-------- 2nd UDG transformation ------------------------------------

-- set the correct attributes, depending on the Elem type
repaint2 :: Elem -> Elem
repaint2 e = case e of 
  R _ -> e
  L (i, N (t,ds,xs)) -> L (i, N (t, chgAttr ds t, map repaint2 xs))
  L (i, E (t,ds,x))  -> L (i, E (t, chgAttr ds t, repaint2 x))

-- change attribute values, depending on the Elem type (as String)
chgAttr :: [Attribute] -> String -> [Attribute]
chgAttr ls e = let 
     ff = setObj ls
  in case e of
    "Cor"   -> setAttr ls [("OBJECT"," Cor "),("COLOR","white"),
                                              ("_GO","triangle")]
    "E2C"   -> setAttr ls [("OBJECT"," E2C "),("COLOR","white"),
                                              ("_GO","triangle")]
    "C2E"   -> setAttr ls [("OBJECT"," C2E "),("COLOR","white"),
                                              ("_GO","triangle")]
    "EC"    -> setAttr ls [("OBJECT","ECand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "CE"    -> setAttr ls [("OBJECT","CEand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "ECand" -> setAttr ls [("OBJECT","ECand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "CEand" -> setAttr ls [("OBJECT","CEand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "R"    -> setAttr ls [("OBJECT",ff "R"),("COLOR","pink"),
                                            ("_GO","rectangle")]
    "rev"  -> setAttr ls [("OBJECT",ff "R"),("COLOR","pink"),
                                            ("_GO","rectangle")]
    "C"    -> setAttr ls [("OBJECT",ff "C"),("COLOR","green"),
                                            ("_GO","rectangle")]
    "cev"  -> setAttr ls [("OBJECT",ff "C"),("COLOR","green"),
                                            ("_GO","rectangle")]
    "E"    -> setAttr ls [("OBJECT",ff "E"),("COLOR","red"),
                                            ("_GO","rectangle")]
    "uev"  -> setAttr ls [("OBJECT",ff "E"),("COLOR","red"),
                                            ("_GO","rectangle")]
    "oco"  -> setAttr ls [("OBJECT",ff "O"),("COLOR","blue"),
                                            ("_GO","rectangle")] 
    "O"    -> setAttr ls [("OBJECT",ff "O"),("COLOR","blue"),
                                            ("_GO","rectangle")]
    "uco"  -> setAttr ls [("OBJECT",ff "U"),("COLOR","orange"),
                                            ("_GO","rectangle")]
    "U"    -> setAttr ls [("OBJECT",ff "U"),("COLOR","orange"),
                                            ("_GO","rectangle")]
    "edge" -> setAttr ls [("_DIR","inverse")]
    _      -> ls 

-- replaces attributes in first list with values in the second    
setAttr :: [Attribute] -> [(String,String)] -> [Attribute]
setAttr ps ls = let 
     qs = [A w | w <- ls]
     f (A (a,b)) (A (c,d)) = a == c
  in nubBy f (qs ++ ps)

-- change the value of the OBJECT attribute (possibly using that 
-- of INFO, if the value of OBJECT is empty)
setObj :: [Attribute] -> String -> String
setObj ls w = let 
     ns = maybe "" id $ extAttr ls "OBJECT"
     hs = dropWhile (\x -> x == ' ') ns
     ks = dropWhile (\x -> x /= ' ' && x /= ':') hs
     ms = dropWhile (\x -> x == ' ' || x == ':') ks
     bs = maybe "" id $ extAttr ls "INFO"
     cs = dropWhile (\x -> x == ' ') bs
     xs = if ms /= "" then ms 
          else if cs /= "" then cs
          else "unspecified"
     n  = length xs
     n1 = 30 - n
  in w ++ " " ++ 
       (if 0 >= n1 then 
              if elem '\n' xs then take 60 xs
              else take 30 xs 
        else let aa = replicate n1 ' ' 
           in take (min 30 $ max 15 $ (length xs) + 5) $ xs ++ aa) 
   
--------------------------------------------------------------------
-------- 3nd UDG transformation ------------------------------------

extLName :: Elem -> String
extLName t = trim $ maybe "" id $ extAName t

-- set the correct attributes, depending on the Elem type
repaint3 :: [String] -> Elem -> Elem
repaint3 ls e = let 
    m = extLName e
  in case e of 
     R _ -> e
     L (i, N (t,ds,xs)) -> L (i, 
                       N (t, chgAttrX ls ds t m, map (repaint3 ls) xs))
     L (i, E (t,ds,x))  -> L (i, E (t, chgAttrX ls ds t m, repaint3 ls x))

-- change attribute values, depending on the Elem type (as String)
chgAttrX :: [String] -> [Attribute] -> String -> String -> [Attribute]
chgAttrX ps ls e m = let 
     ff = setObj ls
     b  = elem m ps
  in case e of
    "Cor"   -> setAttr ls [("OBJECT"," Cor "),("COLOR","white"),
                                              ("_GO","triangle")]
    "E2C"   -> setAttr ls [("OBJECT"," E2C "),("COLOR","white"),
                                              ("_GO","triangle")]
    "C2E"   -> setAttr ls [("OBJECT"," C2E "),("COLOR","white"),
                                              ("_GO","triangle")]
    "EC"    -> setAttr ls [("OBJECT","ECand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "CE"    -> setAttr ls [("OBJECT","CEand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "ECand" -> setAttr ls [("OBJECT","ECand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "CEand" -> setAttr ls [("OBJECT","CEand"),("COLOR","white"),
                                              ("_GO","triangle")]
    "R"    -> setAttr ls [("OBJECT",ff "R"),
               ("COLOR",if b then "purple" else "purple"),("INFO",m),
                                            ("_GO","rectangle")]
    "rev"  -> setAttr ls [("OBJECT",ff "R"),
               ("COLOR",if b then "purple" else "purple"),("INFO",m),
                                            ("_GO","rectangle")]
    "C"    -> setAttr ls [("OBJECT",ff "C"),
                 ("COLOR",if b then "green" else "brown"),("INFO",m),
                                            ("_GO","rectangle")]
    "cev"  -> setAttr ls [("OBJECT",ff "C"),
                 ("COLOR",if b then "green" else "brown"),("INFO",m),
                                            ("_GO","rectangle")]
    "E"    -> setAttr ls [("OBJECT",ff "E"),
                   ("COLOR",if b then "red" else "pink"),("INFO",m),
                                            ("_GO","rectangle")]
    "uev"  -> setAttr ls [("OBJECT",ff "E"),
                   ("COLOR",if b then "red" else "pink"),("INFO",m),
                                            ("_GO","rectangle")]
    "oco"  -> setAttr ls [("OBJECT",ff "O"),
                   ("COLOR",if b then "blue" else "grey"),("INFO",m),
                                            ("_GO","rectangle")] 
    "O"    -> setAttr ls [("OBJECT",ff "O"),
                   ("COLOR",if b then "blue" else "grey"),("INFO",m),
                                            ("_GO","rectangle")]
    "uco"  -> setAttr ls [("OBJECT",ff "U"),
               ("COLOR",if b then "yellow" else "yellow"),("INFO",m),
                                            ("_GO","rectangle")]
    "U"    -> setAttr ls [("OBJECT",ff "U"),
               ("COLOR",if b then "yellow" else "yellow"),("INFO",m),
                                            ("_GO","rectangle")]
    "edge" -> setAttr ls [("_DIR","inverse")]
    _      -> ls 

-------------------------------------------------------------
----------- Auxiliary functions -----------------------------

extAName :: Elem -> Maybe String
extAName e = case e of 
  R _      -> Nothing
  L (_, x) -> extAttr (attributes x) "OBJECT"

extAttr :: [Attribute] -> String -> Maybe String
extAttr ps n = let ws = [y | A (k,y) <- ps, k == n] 
  in case ws of
       []  -> Nothing
       x:_ -> if x == "" then Nothing else Just x

attributes :: ElemProps -> [Attribute]
attributes e = case e of
  N (_,x,_) -> x
  E (_,x,_) -> x

------------------------------------------------------------------
-------- FT analysis function (based on T.FTree) -----------------

analyse :: Graphterm -> T.FTree
analyse [t] = let 
     ls = maybe FNil id $ toTree t 
     b  = nameUnique ls
     xs = ft2event ls
  in if b then ftXtree xs 
          else error "UDType, analyse: behaviour name not unique"

printAna :: T.FTree -> [String] -> IO [()]
printAna ys ns = let 
     ws = T.ft2lg ys
     zs = map (T.NOT . T.AND . (fixAna ns)) [cs | T.AND cs <- ws]
  in printList zs

-- renames conditions out of the control tree
fixAna :: [String] -> [T.Lg String] -> [T.Lg String]
fixAna ls ys = map 
  (\x -> case x of 
      T.OBS w -> if elem w ls then x else T.AOBS w
      _     -> x) ys

-------------------------------------------------------------------
-------- translation from Elem to FTreeX (a standard Tree structure)

data FT a = FNil | FT a [FT a]
  deriving (Ord, Eq, Show)

type FTreeX = FT (String,String,String)

toTree :: Elem -> Maybe FTreeX
toTree e = case e of 
  R _     -> Nothing
  L (x,y) -> case y of 
    E (_, _, w) -> toTree w
    N (t,as,ws) -> do 
       zs <- mapM toTree ws
       d  <- extAName e
       return $ FT (t, trim d, x) zs

behTags :: [String]
behTags = ["C","cev","E","uev","O","oco","U","uco","R","rev"]

nameList :: FTreeX -> [String]
nameList t = case t of
  FT (a,b,_) ls -> if elem a behTags 
                   then b:(concat $ map nameList ls)
                   else concat $ map nameList ls
   
nameUnique :: FTreeX -> Bool
nameUnique t = let
    ls = nameList t
 in (length $ nub ls) == length ls

------------------------------------------------------------------
------- translate from FTreeX to EventTree -----------------------

ft2event :: FTreeX -> EventTree
ft2event t = case t of 
  FT (c,b,_) [] -> EC (ft2ev c b) []
  FT (c,b,_) [FT (_,_,_) []] -> trace (show t) 
           $ error ("ft2event: type error 1; "
               ++ "\n inner node should be a gate, cannot be a leaf")
  FT (c,b,_) [FT (_,"ECand",_) fs@(_:_)] -> 
                                 EC (ft2ev c b) (map ft2cond fs)
  FT (c,b,_) [FT (_,"C2E",_) [f]] -> EC (ft2ev c b) [ft2cond f]
  FT (c,b,_) [FT (_,"C2E",_) _] -> trace (show t) 
           $ error ("ft2event: type error 2; "
               ++ "\n C2E gate cannot have more than an input node")
  FT (c,b,_) [FT (_,_,_) _] -> trace (show t) 
           $ error ("ft2event: type error 3; "
               ++ "\n not a fault tree: inner node should be "
               ++ "a node, either C2E or ECand")
  _ -> trace (show t) $ error ("ft2event: type error 4; "
            ++ "\n behaviour node cannot have more than an input node")

ft2ev :: String -> String -> Event
ft2ev t s = case t of 
  "uev" -> NCont s
  "cev" -> Cont s
  "rev" -> Rare s
  "E" -> NCont s
  "C" -> Cont s
  "R" -> Rare s
  _ -> trace (show t) $ error ("ft2ev: type error 5; "
                   ++ "\n node is not an event")

ft2co :: String -> String -> Cond
ft2co t s = case t of 
  "oco" -> Obs s
  "uco" -> Nobs s
  "O" -> Obs s
  "U" -> Nobs s
  _ -> trace (show t) $ error ("ft2ev: type error 6; "
                 ++ "\n node is not a condition")

ft2cond :: FTreeX -> SimpleCond
ft2cond t = case t of 
  FT (c,b,_) [FT (_,_,_) []] -> trace (show t) 
           $ error ("ft2event: type error 7; "
               ++ "\n inner node should be a gate, cannot be a leaf")
  FT (c,b,_) [FT (_,"Cor",_) fs] -> 
      COr (ft2co c b) (map ft2cond fs)
  FT (c,b,_) [FT (_,"E2C",_) ffs@(f:fs)] -> 
      case fs of 
         [] -> CE (ft2co c b) (ft2event f)
         _  -> trace (show t) $ error ("ft2event: type error 8; "
                ++ "\n gate E2C cannot have more than an input node")
  FT (c,b,_) [FT (_,"CEand",_) ffs@(f:fs)] -> let 
      ps = case fs of 
         [] -> ft2event f
         _  -> let 
             ws = [x | x <- ffs, elem (tyFT x) 
                              ["cev","uev","rev","E","C","R"]]
           in case ws of 
              [FT (c,b,_) []] -> let zs = delete (head ws) ffs
                   in EC (ft2ev c b) $ map ft2cond zs
              [FT (c,b,_) _] -> trace (show t)  
                 $ error ("ft2event: type error 9; "
                   ++ "\n event input of CEand gate with input " 
                      ++ "conditions \n must be a leaf node") 
              _ -> trace (show t) $ error ("ft2event: type error 10; "
                   ++ "CEand gate cannot have more than an event input") 
    in CE (ft2co c b) ps
  FT (c,b,_) [] -> CE (ft2co c b) NilEv
  FT (c,b,_) [FT (_,_,_) fs@(_:_)] -> trace (show t) 
           $ error ("ft2event: type error 12; "
               ++ "\n not a fault tree: inner node should be "
               ++ "a node, either E2C, CEand or Cor")
  _ -> trace (show t) $ error ("ft2event: type error 13; "
            ++ "\n behaviour node cannot have more than a child node")
      
tyFT :: FTreeX -> String  -- extract Elem type
tyFT (FT (t,_,_) _) = t

--------------------------------------------------------------------
------- translate from EventTree to T.FTree ------------------------

ftXtree :: EventTree -> T.FTree
ftXtree t = case t of
  EC (Cont x) fs  -> T.Tr (T.mkTr T.Y x) (T.Nd T.GAnd $ map scXtree fs)
  EC (NCont x) fs -> T.Tr (T.mkTr T.N x) (T.Nd T.GAnd $ map scXtree fs)
  _ -> trace (show t) $ error "ftXtree: type error"

scXtree :: SimpleCond -> T.FTree
scXtree e = case e of 
  CE p (EC t fs) -> let 
         np = case p of
            Obs s  -> T.Place T.Y s
            Nobs s -> T.Place T.N s
         nt = case t of
            Cont s  -> T.Trans T.Y s
            NCont s -> T.Trans T.N s
            Rare s  -> T.Trans T.I s
     in T.Tr (T.Event np nt) (T.Nd T.GAnd $ map scXtree fs)
  CE p NilEv -> let 
         np = case p of
            Obs s  -> T.mkPl T.Y s
            Nobs s -> T.mkPl T.N s
     in T.Tr np (T.Nd T.GAnd [])
  COr p fs -> let 
         np = case p of
            Obs s  -> T.mkPl T.Y s
            Nobs s -> T.mkPl T.N s
     in T.Tr np (T.Nd T.GOr $ map scXtree fs)
             
--------------------------------------------------------------------

